from numpy import *
import binascii, struct
from pyec.util.TernaryString import TernaryString


class BayesRandomizer(object):
   def __call__(self, network):
      return random.randn(len(network.variables))

class BinaryRandomizer(BayesRandomizer):
   def __call__(self, network):
      numBytes = int(ceil(len(network.variables) / 8.0))
      numFull  = len(network.variables) / 8
      initial = ''
      if numBytes != numFull:
         extra = len(network.variables) % 8
         initMask = 0
         for i in xrange(extra):
            initMask <<= 1
            initMask |= 1
         initial = struct.pack('B',initMask)
            
      base = long(binascii.hexlify(random.bytes(numBytes)), 16)
      known = long(binascii.hexlify(initial + '\xff'*numFull), 16)
      return TernaryString(base, known)

class MultinomialRandomizer(BayesRandomizer):
   def __call__(self, network):
      cats = network.variables[0].categories
      ret = zeros(len(cats))
      for variable in network.variables:
         ret[variable.index] = random.randint(0, len(cats[variable.index])) + 1
      return ret


class BayesSampler(object):
   pass
   
   
class DAGSampler(BayesSampler):
   def __call__(self, network, initial = None):
      # order the variables in a DAG
      # call the variables one by one
      if initial is None:
         ret = network.randomize()
      else:
         ret = initial
      for variable in network.variables:
         if initial is None or not variable.isset(initial):
            ret = variable(ret)
      return ret


class GibbsSampler(BayesSampler):
   def __init__(self, equilibrium):
      self.equilibrium = equilibrium

   def __call__(self, network, initial = None):
   
      # get a random state
      ret = network.randomize()
      numNotSet = len(network.variables)
      if initial is not None:
         numNotSet = 0
         for variable in network.variables:
            if variable.isset(initial):
               ret[variable.index] = initial[variable.index]
         numNotSet += 1
      if numNotSet == 0:
         return initial
      if numNotSet == 1:
         for variable in network.variables:
            if initial is None or not variable.isset(initial):
               ret = variable.marginal(ret, network)
         return ret
      for i in xrange(self.equilibrium):
         index = random.randint(0, len(network.variables))
         while network.variables[index].isset(ret):
            index = random.randint(0, len(network.variables))
         ret = network.variables[index].marginal(ret, network)
      return ret
      
class MixtureSampler(BayesSampler):
   def __init__(self, segment):
      self.segment = segment
      
   def __call__(self, network, initial = None):
      from pyec.db.models import Point
      if initial is None:
         key = TernaryString(0L,0L)
         bits = 0
      else:
         key = initial
         bits = 0
         for variable in network.variables:
            if not variable.isset(initial):
               break
            bits += 1
      # TODO - convert initial to binary format for sampling the Partition tree
      return Point.objects.sampleConditional(self.segment, key, bits, self.segment.config)
